package template;

import sleep.parser.*;
import sleep.engine.*;
import sleep.interfaces.*;
import sleep.taint.*;
import sleep.runtime.*;
import sleep.bridges.*;
import sleep.error.*;

import java.io.*;
import java.util.*;

/** This class is an implementation of the Sleep template engine.  The template engine constructs
 *  an executable template that consists of strings to print and code to evaluate.</p>
 * 
 *  Valid templates include:
 *
 *  &lt;% <em>expression</em> &gt;%>
 *  &lt;?sleep <em>statement(s)</em> ?&gt;
 */
public class TemplateEngine implements Loadable, Function
{
   public void scriptLoaded(ScriptInstance script)
   {
      script.getScriptEnvironment().getEnvironment().put("&display", this);
   }

   public void scriptUnloaded(ScriptInstance script) { }

   public Scalar evaluate(String name, ScriptInstance script, Stack args)
   {
      File f = BridgeUtilities.getFile(args, script);
      
      if (!f.exists())
         throw new IllegalArgumentException(name + ": Hey! " + f + " does not exist...");
      
      Template temp = null;
      synchronized (templateCache)
      {
         temp = (Template)templateCache.get(f);
         if (temp == null || !temp.isValid())
         {
            try
            {
               System.err.println("[Moconti] " + f + " " + (temp == null ? "did not exist" : "changed") + ", loading template");

               temp = parseTemplate(f, readFully(new FileReader(f)));
               templateCache.put(f, temp);
            }
            catch (Exception ex)
            {
               throw new RuntimeException(ex);
            }
         }
      }

      if (temp != null)
      {
         temp.evaluate(script, args);
      }
      return SleepUtils.getEmptyScalar();
   }

   protected Map templateCache = new HashMap();

   /** parses the template code */
   protected Block parseCode(String name, int lineNo, ImportManager imports, String code) throws YourCodeSucksException
   {
      Parser temp = new Parser(name, "", imports); 

      if (TaintUtils.isTaintMode())    
      {    
         temp.setCodeFactory(new TaintModeGeneratedSteps());    
      }    

      temp.parse(new StringIterator(code, lineNo));
      return temp.getRunnableBlock();
   }

   /** parses a Sleep template file.  Lots of fun for the whole family */
   public Template parseTemplate(File name, String code) throws YourCodeSucksException
   {
      ImportManager imports = new ImportManager();
      Template rv = new Template(name);

      StringBuffer text  = new StringBuffer(1024);
      StringBuffer codez = new StringBuffer(1024);
      int          lno;

      StringIterator i = new StringIterator(code);
      if (i.hasNext()) { i.next(); } /* skip the initial newline at the beginning */

      while (i.hasNext())
      {
         if (i.isNextString("<%"))
         {
            rv.addString(text.toString());
            text.setLength(0);

            i.next(2);

            codez.setLength(0);
            lno   = i.getLineNumber();

            while (i.hasNext() && !i.isNextString("%>"))
            {
               codez.append(i.next());                
            }

            i.next(2);

            rv.addStatements(parseCode(name.toString(), lno, imports, "print(" + codez + ");"));
         }
         else if (i.isNextString("<?sleep"))
         {
            rv.addString(text.toString());
            text.setLength(0);

            i.next(7);

            codez.setLength(0);
            lno   = i.getLineNumber();

            while (i.hasNext() && !i.isNextString("?>"))
            {
               codez.append(i.next());                
            }

            i.next(2);

            rv.addStatements(parseCode(name.toString(), lno, imports, codez.toString()));
         }
         else
         {
            text.append(i.next());
         }
      }

      rv.addString(text.toString());

      return rv;
   }
 
    private static String readFully(Reader reader)
    {
        StringBuffer code = new StringBuffer(8192);

        try
        {
           BufferedReader in = new BufferedReader(reader);
           String s = in.readLine();
           while (s != null)  
           {
               code.append("\n");
               code.append(s);
               s = in.readLine();
           }

           in.close();
        }
        catch (Exception ex) { }

        return code.toString();
    }

   public static void main(String args[]) throws Exception
   {
       ScriptLoader loader = new ScriptLoader();
       ScriptInstance script = loader.loadScript("phear", "", null);
       script.addWarningWatcher(new RuntimeWarningWatcher()
       {
          public void processScriptWarning(ScriptWarning warning)
          {
             System.out.println(warning.toString());
          }
       });

       Template temp = new TemplateEngine().parseTemplate(new File(args[0]), readFully(new FileReader(args[0])));        
       temp.evaluate(script, new Stack());
   }
}


